#ifndef H_BASIC_OBJECT 
#define H_BASIC_OBJECT

#include "Point.h"

/**
  class providing basic functionality for all other objects
  note that any object is ancestor of BasicObject
  */

class BasicObject{
    public:

      /**
        each object have unique id, so this is the best
        way to identify it
        */          
      int id;

      /**
        radius of object - object is circle
        */
      double radius;

      /**
        rotation of object. For some objects used only
        for graphics, for others play important role
        */
      double angle;

      /**
        position of object
        */
      Point position;

      /**
        tells whether is object visible. If you see this flag false,
        it can be only if this is your Ship (e.g. any other objects
        in worldarray have always visible=true (not visible object are
        not passed to you
        */
      int visible;

      /**
        internal flag - alive=false means to kill object in next
        sweeping process
        */
      int alive;

      /**
        number of owning player -
        -1 for world,
        0.. for players
        */
      int owner;

    private:
      /**
        incremented by each creation of object
        */
      static int last_id;
    public:

      /**
        constructor, clears all variables
        */
      BasicObject();

      virtual ~BasicObject();

      /**
        notify object that world is going to make new time quantum
        */
      virtual void timeNotify(double timeDelta);

      /**
        before next frame, clear all temporary variables
        */
      virtual void clean();

      /**
        copy BasicObject's properties to another BasicObject
        */
      static void copyData(const BasicObject* from, BasicObject* to);

      /**
        serialize objects data, not putting object's mark
        */
      static void _serialize(std::vector<unsigned char>& destination,
              BasicObject* ptr);
      /**
        deserialize objects data
        */
      static void _deserialize(std::vector<unsigned char>::iterator&
              it, BasicObject* ptr);


      /**
        virtual function - each object implements own serialization
        */
      virtual void serialize(std::vector<unsigned char>&destination)=0;

      /**
        virtual function - tell graphics to draw this object
        */
      virtual void draw();

      /**
        return type of object - commonly used for casting
        */
      virtual int getType()=0;

};

#endif
