#ifndef H_COLLIDER
#define H_COLLIDER

#include <vector>
#include "MassObject.h"
#include <utility>
#include <set>
#include "Preferences.h"

class World;

class Collider{
    public:
    /**
      constructor
      */
    Collider();

    /**
      initialize collider from preferences and world size
      (needed by bucketing)
      */
    void initialize(World* world);

    /**
      clear bucketing state
      */
    void clear();

    /**
      insert into bucketing engine Mass objects and
      automatically performs collisions 
      (efectively saving computations)
      */
    // {{{ insertAndCollideMassObjects
    template <typename InputIterator>
    void insertAndCollideMassObjects(InputIterator begin, InputIterator end){
        while (begin!=end){
            insertAndCollideMassObject(*begin);
            ++begin;
        }
    }
    // }}}

    /**
      collide massless objects with mass objects inserted by
      insertAndCollideMassObjects()
      */
    // {{{ collideMasslessObjects
    template <typename InputIterator>
    void collideMasslessObjects(InputIterator begin, InputIterator end){
        while (begin!=end){
            collideMasslessObject(*begin);
            ++begin;
        }
    }
    // }}}

    /**
      return number of buckets on X axis
      */
    int getBucketsX() const;

    /**
      return number of buckets on Y axis
      */
    int getBucketsY() const;

    /**
      return number of objects in specific bucket
      */
    int getDensity(const int x,const int y) const;

    /**
      return size of bucket in pixels
      */
    int getBucketSize() const;

    Point getSpawnPoint();

    private:

    void initializeSpawnData();

    
    /**
      performs very precise detection of collision
      */
    void testMassMassCollision(MassObject* object1, MassObject* object2);

    /**
      performs very precise detection of collision
      */
    void testMassMasslessCollision(MassObject* object1, BasicObject* object2);

    /**
      actual testing routine
      p - position difference
      d - speed difference
      r - radius
      return true if objects collide
      */
    int testCollision(const Point p,const Point d,const double r);

    void insertAndCollideMassObject(MassObject* const object);

    void collideMasslessObject(BasicObject* const object);


    /**
      insert collision buckets for point and radius
      */
    void getCollisionBuckets(const Point pos, const double radius);

    /**
      for given point and radius and xPosition of actual bucket
      and position of point's normal bucket compute minimum Y bucket
      */
    void getBucketingY(const Point position, const double radius,
            const int xZero, const int xPosition, int& min, int& max) const;


    /**
      actual bucketing data
      */
    std::vector<std::vector<std::vector<MassObject*> > >
         bucketList;

    /**
      temporary collision data, it is 'global' only because
      of big overhead to alloc/dealloc
      */
    std::vector<std::vector<MassObject*> *> collisions;

    std::vector<MassObject*> haveCollided;

    /**
      number of buckets on X axis
      */
    int bucketsX;

    /**
      number of buckets on Y axis
      */
    int bucketsY;

    /**
      size of bucket
      */
    double bucketSize;

    std::vector<double> spawnWeight;
    int spawnPoints;
    World* world;
};

#endif
