#ifndef H_SHIP
#define H_SHIP

#include "MassObject.h"
#include "PlayerResponse.h"
#include <vector>
#include "WeaponSlot.h"

class World;

#define WEAPON_SLOTS 3

class Ship: public MassObject{
    public:
        /**
          we must provide explicit copy operators
          */
        Ship& operator=(const Ship& ship);

        /**
          explicit copy constructor
          */
        Ship(const Ship& ship);

        /**
          player number
          note that it shoud be equal to BasicObject's owner
          */
        int player;

        /**
          ship number,
          each player's ships are numbered 0...(n-1)
          */
        int ship;

        /**
          ship's live
          default value is 100
          */
        double live;

        /**
          constructor
          */
        Ship();

	virtual ~Ship();

        /**
          internal function used for serializing to bytestream
          */
        virtual void serialize(std::vector<unsigned char>& destination);

        /**
          internal function used to deserialize bytestream
          */
        static Ship* deserialize(std::vector<unsigned char>::iterator&
                it);

        /**
          return type of object - Types::ship
          */
        virtual int getType();

        /**
          tell ship that it will be destroyed - it will add explosion
          to world
          */
        virtual void destroy(World* world);

        /**
          notify ship that new timeQunatum started
          */
        virtual void timeNotify(double timeDelta);

        /**
          notify ship that it received damage from impact and
          explosion,
          pointer to world is supplied to do some ohter tasks as
          adding explosions, etc.
          */
        virtual void doDamage(double impactDamage,double
                explosionDamage, int from, World* world);

        /**
          clean temporary data before next frame
          */
        virtual void clean();

        /**
          draw ship to screen
          */
        virtual void draw();

        /**
          copy ship data - interlanny used by constructors
          */
        static void copyData(const Ship* from, Ship* to);

        /**
          compute player response
          */
        void doPlayerMove(World* world);

        /**
          clear ship's data for other players - internally used
          */
        void clearForOthers();

        /**
          tell whether ship is immutable - each ship is for a few
          while after spawning immutable
          */
        int isImmutable();

        /**
          wepoan slots - each one represents one weapon
          */
        WeaponSlot* weaponSlots[WEAPON_SLOTS];

        /**
          called when ship received autodestruct command
          */
        void autoDestruct(World* world);
	
	/**
          player response as received over line
          */
        PlayerResponse response;

    private:
        /**
          tells remaining time of inMutability
          */
        double nodamageCountdown;

        /**
          internal information - last damage on ship,
          used for calculating final explosion
          */
        double lastDamage;

        /**
          internal information - who last was doing damage on ship,
          we will add frags to him if he killed ship
          */
        int lastDamageFrom;

    public:
    /**
      initialize all static variables
      */
    static void initialize();

    /**
      tells whether is ship initialized. only for asserts
      */
    static int initialized;

    /**
      default live of ship as readed from proboj.ini
      */

    static double defaultLive;
    /**
      default immutable time as readed from ini
      */
    static double defaultNodamageTime;

    /**
      max speed of ship. Accelerating over this speed have no
      effect
      */
    static double maxSpeed;

    /**
      max acceleration of ships, if it uses full jets
      */
    static double maxAcceleration;

    /**
      max rotational acceleration,
      very useful for navigating,
      @see Utils::getShipRotationalJets
      */
    static double maxRotationalAcceleration;

    /**
      standard mass of ship. currently all ships have this mass
      */
    static double defaultMass;

    /**
      standard radius of ship. currently all ships have this radius
      */
    static double defaultRadius;

    /**
      standard angular mass, currently all ships have this radius
      */
    static double defaultRotationalMass;

    /**
      if ship receives command to autoDestruct, set lastDamage to this
      value
      */
    static double autoDestructLastDamage;

    /**
      final explosion is calculated as
      explosionStrength+lastDamage*lastDamageExplosionFactor
      */
    static double lastDamageExplosionFactor;

    /**
      @see lastDamageExplosionFactor
      */
    static double explosionStrength;
};
#endif
