
#include "Collider.h"
#include "Preferences.h"
#include "Debug.h"
#include "Macro.h"
#include "Assert.h"
#include "Point.h"
#include <cmath>
#include "MyMath.h"
#include "Random.h"
#include "CollisionSolver.h"
#include "Graphics.h"
#include "World.h"

// {{{ Collider OK
Collider::Collider():bucketsX(0),bucketsY(0),bucketSize(1){
}
// }}}

// {{{ initialize OK
void Collider::initialize(World* world){
  bucketSize=Preferences::_getDouble("/collider/bucketSize");
  bucketsX=1+(int)(world->width/bucketSize);
  bucketsY=1+(int)(world->height/bucketSize);
  bucketList.resize(bucketsX);
  FORSIZE(q,bucketList) bucketList[q].resize(bucketsY);
  initializeSpawnData();
  this->world=world;
  this->clear();
}
// }}}

// {{{ clear OK
void Collider::clear(){
    assert(bucketsX==(int)bucketList.size(),"internal inconsistency "
            "in collider");
    assert(bucketsX>0,"have you initialized collider?");
    assert(bucketsY>0,"have you initialized collider?");

    FOR(q,bucketsX) {
        assert(bucketsY==(int)bucketList[0].size(),"");

        FOR(w,bucketsY)
            bucketList[q][w].resize(0);
    }
}
// }}}

// {{{ getBucketingY OK
void Collider::getBucketingY(const Point position, const double
        radius,const int xZero, const int xPosition,int& min, int& max) const{

    if (xZero==xPosition) {
        min=(int) ((position.y-radius)/bucketSize);
        max=(int) ((position.y+radius)/bucketSize);
        return;
    }

    double dx;

    if (xZero>xPosition) {
        dx=bucketSize*(xPosition+1)-position.x;
    } else {
        dx=bucketSize*xPosition-position.x;
    }

    double squareLength=radius*radius-dx*dx;
 //   assert(squareLength>=0,"Collider - internal error");

    double dy=sqrt(squareLength);
    min=(int) ((position.y-dy)/bucketSize);
    max=(int) ((position.y+dy)/bucketSize);
    
}
// }}}

// {{{ insertAndCollideMassObject OK
void Collider::insertAndCollideMassObject(MassObject* const object){

    Point pos[2]={object->position,object->getFlipPosition()};

    if ((pos[0]-pos[1]).length()>bucketSize)
        Errors::_addError("Too fast object, should incorrectly detect"
               " collisions",Errors::WARNING);

    collisions.resize(0);
    getCollisionBuckets(pos[0],object->radius);
    getCollisionBuckets(pos[1],object->radius);
    sort(collisions.begin(),collisions.end());

    std::vector<std::vector<MassObject*>*>::iterator 
            endIterator=std::unique(collisions.begin(),collisions.end());
    std::vector<std::vector<MassObject*>*>::iterator
            startIterator=collisions.begin();

    haveCollided.resize(0);

    while (startIterator!=endIterator){
           std::vector<MassObject*>::iterator 
               innerEndIterator=(*startIterator)->end();
           std::vector<MassObject*>::iterator 
               innerBeginIterator=(*startIterator)->begin();

           while (innerBeginIterator!=innerEndIterator) {
               int ok=1;
               FOREACH(it,haveCollided)
                   if (*it==*innerBeginIterator) ok=0;

               if (ok) {
                   testMassMassCollision(*innerBeginIterator,object);
                   haveCollided.push_back(*innerBeginIterator);
                   }
               ++innerBeginIterator;
           }
          (*startIterator)->push_back(object);
        ++startIterator;
    }
}
// }}}

// {{{ collideMasslessObject OK
void Collider::collideMasslessObject(BasicObject* const object){
    collisions.resize(0);
    getCollisionBuckets(object->position,object->radius);

    std::vector<std::vector<MassObject*>*>::iterator 
            endIterator=collisions.end();
    std::vector<std::vector<MassObject*>*>::iterator
            startIterator=collisions.begin();

    haveCollided.resize(0);

    while (startIterator!=endIterator){
           std::vector<MassObject*>::iterator 
               innerEndIterator=(*startIterator)->end();
           std::vector<MassObject*>::iterator 
               innerBeginIterator=(*startIterator)->begin();

           while (innerBeginIterator!=innerEndIterator) {
               int ok=1;
               FOREACH(it,haveCollided)
                   if (*it==*innerBeginIterator) ok=0;

               if (ok) {
                   testMassMasslessCollision(*innerBeginIterator,object);
                   haveCollided.push_back(*innerBeginIterator);
                }
               ++innerBeginIterator;
           }

        ++startIterator;
    }
}
// }}}

// {{{ getCollisionBuckets OK
void Collider::getCollisionBuckets(const Point pos,const double r){
    int start=(int) ((pos.x-r)/bucketSize);
    int zero=(int) (pos.x/bucketSize);
    int end=(int) ((pos.x+r)/bucketSize);

    if (start<0) start=0;
    if (end>=bucketsX) end=bucketsX-1;

    for (int i=start; i<=end; i++){
        int bottom;
        int up;
        getBucketingY(pos,r,zero,i,bottom,up);
        if (bottom<0) bottom=0; // clip region
        if (up>=bucketsY) up=bucketsY-1;

        for (int k=bottom; k<=up; k++)
            collisions.push_back(&bucketList[i][k]);
    }
}
// }}}

// {{{ testMassMassCollision OK
void Collider::testMassMassCollision(MassObject* object1, MassObject* object2){
    Point p=object1->position-object2->position;
    Point d=object1->getFlipPosition()-object2->getFlipPosition()-p;
    if (testCollision(p,d,object1->radius+object2->radius)){
        CollisionSolver::doMassMassCollision(object1,object2,world);
    }
}
// }}}

// {{{ testMassMasslessCollision OK
void Collider::testMassMasslessCollision(MassObject* object1,
        BasicObject* object2){
    Point p=object1->position-object2->position;
    Point d=object1->getFlipPosition()-object1->position;
    if (testCollision(p,d,object1->radius+object2->radius)){
        CollisionSolver::doMassMasslessCollision(object1,object2,world);
    }
}
// }}}

// {{{ testCollision OK
int Collider::testCollision(const Point p,const Point d,const double r){
    double t=0;
    if (d.length2()>0.1) {
        t=-(p*d)/d.length2();
    }
    if (t<0) t=0;
    if (t>1) t=1;

    if ((p+t*d).length()< r) return 1;
    return 0;
}
// }}}



// {{{ getBucketsX OK
int Collider::getBucketsX() const{
    return bucketsX;
}
// }}}

// {{{ getBucketsY OK
int Collider::getBucketsY() const{
    return bucketsY;
}
// }}}

// {{{ getDensity OK
int Collider::getDensity(const int x, const int y) const{
    assert(x>=0 && y>=0 && x<bucketsX && y<bucketsY,"getDensity, wrong"
           " parameter");
    return bucketList[x][y].size();
}
// }}}

// {{{ getBucketSize OK
int Collider::getBucketSize() const{
    return (int) bucketSize;
}
// }}}

// {{{ initializeSpawnData OK
void Collider::initializeSpawnData(){
    spawnPoints=Preferences::_getInt(
            "/collider/spawnData/randomPoints");
    spawnWeight.resize(5);
    for (int i=0;i<5;i++)

    spawnWeight[i]=Preferences::_getDouble(
            "/collider/spawnData/weight"+
            MyStringUtils::intToString(i));
}
// }}}

// {{{ getSpawnPoint OK
Point Collider::getSpawnPoint(){
    /*
    std::vector<std::vector<int> > data(bucketsX);

    FOR(x,bucketsX){
        data[x].resize(bucketsY);
        FOR(y,bucketsY)
         data[x][y]=getDensity(x,y);
    }
    */

    std::pair<double, std::pair<int,int> > guess,
        best;

    best.first=1e10;

    FOR(q, spawnPoints){
        double goodness = Random::getDouble(50) ;
        int x = Random::getInt(bucketsX-3)+1;
        int y = Random::getInt(bucketsY-3)+1;

        for (int xx=-2; xx<=2; xx++)
            for (int yy=-2 ; yy<=2 ; yy++)
                 if (xx+x >=0 && xx+x < bucketsX &&
                     yy+y >=0 && yy+y < bucketsY)
                goodness += getDensity(xx+x,yy+y)* //data[xx+x][yy+y] *
                    spawnWeight[std::abs(xx) + std::abs(yy)]
                    * (1 + 0.1 * Random::getDouble());

        guess=std::make_pair(goodness,
                std::make_pair(x, y));
//        printf("spawn guess %d %d  %lf\n",x,y,goodness);
        best=std::min(best, guess);
    }
    return
        Point(bucketSize * (best.second.first + Random::getCenteredDouble(0.5, 0.3)),
              bucketSize * (best.second.second+ Random::getCenteredDouble(0.5, 0.3)));
}
// }}}
