#include "PlayerStatistics.h"
#include "Errors.h"
#include "Assert.h"
#include "Macro.h"
#include "Serializer.h"
#include "MyStringUtils.h"
#include "Preferences.h"

// {{{ initialize
void PlayerStatistics::initialize(int playerCount){
    playerCount++;
    kills.resize(playerCount);
    deaths.resize(playerCount);
    totalDamage.resize(playerCount);
    shipDamage.resize(playerCount);
    weaponsFired.resize(playerCount);
    receivedShipDamage.resize(playerCount);
    for (int q=0; q<playerCount; q++)
        receivedShipDamage[q].resize(playerCount);

    this->playerCount=playerCount;
}
// }}}

// {{{ addDeath
void PlayerStatistics::addDeath(int player){
    player++;
    assert(player<playerCount,"");
    deaths[player]++;
}
// }}}

// {{{ addKill
void PlayerStatistics::addKill(int player){
    player++;
    assert(player<playerCount,"");
    kills[player]++;
}
// }}}

// {{{ addShoot
void PlayerStatistics::addShoot(int player){
    player++;
    assert(player<playerCount,"");
    weaponsFired[player]++;
}
// }}}

// {{{ addDamage 
void PlayerStatistics::addDamage(int player,double totalDamage,
        double shipDamage, int to){
    player++;
    assert(player<playerCount,"");
    to++;
    assert(to<playerCount,"");

    this->totalDamage[player]+=totalDamage;
    this->shipDamage[player]+=shipDamage;
    this->receivedShipDamage[player][to]+=shipDamage;
}
// }}}

// {{{ write OK
void PlayerStatistics::write(){
    printf("kills\n");
    FOR(q,playerCount) printf("%d ",kills[q]);
    printf("\n");

    printf("deaths\n");
    FOR(q,playerCount) printf("%d ",deaths[q]);
    printf("\n");

    printf("total damage\n");
    FOR(q,playerCount) printf("%5.1f ",totalDamage[q]);
    printf("\n");


    printf("ship damage\n");
    FOR(q,playerCount) printf("%5.1f ",shipDamage[q]);
    printf("\n");

    printf("ship damage for pairs:\n");
    FOR(q,playerCount) if (q) {
        printf("%d :   ",q-1);
        FOR(w,playerCount) if (w)
          printf("%5.1f ",receivedShipDamage[q][w]);
        printf("\n");


    }
    printf("\n-------------- TOTAL SCORE: \n");
    FOR(q,playerCount-1) printf("%d: %.2lf\n",q, getScore(q));

}
// }}}

// {{{ getScore
double PlayerStatistics::getScore(int player){
    assert(player>=0,"");
    player++;
    assert(player<=playerCount,"");
    double score=0;
    FOR(q,playerCount) if (q)
         score+=2.5*receivedShipDamage[player][q];

    FOR(q,playerCount) if(q) score-=2*receivedShipDamage[q][player];
    score-=2*receivedShipDamage[player][player];

    score+=30*kills[player];
    score-=30*deaths[player];
    score+=0.3*totalDamage[player];
    return score;
}
// }}}

// {{{ serialize
void PlayerStatistics::serialize(std::vector<unsigned char>&
        destination){
    Serializer::serializeInt(destination,playerCount);
    FOR(q,playerCount){
        Serializer::serializeInt(destination, kills[q]);
        Serializer::serializeInt(destination, deaths[q]);
        Serializer::serializeFloat(destination, totalDamage[q]);
        Serializer::serializeFloat(destination, shipDamage[q]);
        FOR(w,playerCount)
          Serializer::serializeFloat(destination,
                  receivedShipDamage[q][w]);
    }   
    Serializer::serializeChar(destination,47);
}
// }}}

// {{{ deserialize
void PlayerStatistics::deserialize(std::vector<unsigned
        char>::iterator& it){
    playerCount=Serializer::deserializeInt(it);
    initialize(playerCount-1);

    FOR(q,playerCount) {
        kills[q]=Serializer::deserializeInt(it);
        deaths[q]=Serializer::deserializeInt(it);

        totalDamage[q]=Serializer::deserializeFloat(it);
        shipDamage[q]=Serializer::deserializeFloat(it);



        FOR(w,playerCount)
             receivedShipDamage[q][w] =
             Serializer::deserializeFloat(it);
    }
    assert(Serializer::deserializeChar(it)==47,"data corrupted");    
}
// }}}

// {{{ saveToFile
void PlayerStatistics::save(Preferences& data, std::string key){
  FOR(q,playerCount-1)
      data.setDouble(key+"/"+MyStringUtils::intToString(q),
                getScore(q));

}
// }}}
