#include "Ship.h"
#include "Types.h"
#include "Serializer.h"
#include "Preferences.h"
#include "Assert.h"
#include "World.h"
#include "PlasmaBall.h"
#include "PlasmaWeaponSlot.h"
#include "CannonWeaponSlot.h"
#include "MissileWeaponSlot.h"
#include "ObjectGraphics.h"

#include "stdio.h"

int Ship::initialized=0;
double Ship::defaultLive;
double Ship::defaultNodamageTime;
double Ship::maxSpeed;
double Ship::maxAcceleration;
double Ship::maxRotationalAcceleration;
double Ship::defaultMass;
double Ship::defaultRadius;
double Ship::defaultRotationalMass;
double Ship::explosionStrength;
double Ship::lastDamageExplosionFactor;
double Ship::autoDestructLastDamage;

int Ship::isImmutable(){
    return nodamageCountdown>0;
}

Ship::~Ship(){
//    fprintf(stderr,"delete ship\n");
 FOR(q,WEAPON_SLOTS)
      delete weaponSlots[q];
}

Ship& Ship::operator=(const Ship& ship){
    copyData(&ship,this);        

    return *this;
}

void Ship::copyData(const Ship* from, Ship* to){
    MassObject::copyData(from,to);
    to->player=from->player;
    to->ship=from->ship;
    to->live=from->live;
    to->nodamageCountdown=from->nodamageCountdown;
    to->lastDamage=from->lastDamage;

    FOR(q,WEAPON_SLOTS) {
             delete to->weaponSlots[q];
             to->weaponSlots[q]=NULL;

             if (from->weaponSlots[q]!=NULL)
               to->weaponSlots[q]=from->weaponSlots[q]->clone();
    }

}

Ship::Ship(const Ship& ship):MassObject(){
    FOR(q,WEAPON_SLOTS) weaponSlots[q]=NULL;

    copyData(&ship,this);
        
}


void Ship::clearForOthers(){
    live=0; nodamageCountdown=0; lastDamage=0;
    lastDamageFrom=0;
    FOR(q,WEAPON_SLOTS) {delete weaponSlots[q]; weaponSlots[q]=NULL;}
}

// {{{ clean
void Ship::clean(){
    MassObject::clean();
    lastDamage=0;
    lastDamageFrom=-1;
}
// }}}

// {{{ getType
int Ship::getType(){
    return Types::ship;
}
// }}}

// {{{ serialize
void Ship::serialize(std::vector<unsigned char>&
        destination){

    Serializer::serializeChar(destination,Types::ship);
    MassObject::_serialize(destination,this);
    Serializer::serializeChar(destination,player);
    Serializer::serializeChar(destination,ship);


    Serializer::serializeFloat(destination,live);
    Serializer::serializeFloat(destination,nodamageCountdown);

    FOR(q,WEAPON_SLOTS)
        weaponSlots[q]->serialize(destination);
}
// }}}

// {{{ deserialize
Ship* Ship::deserialize(std::vector<unsigned
        char>::iterator & it){
    Ship* ptr=new Ship();
    MassObject::_deserialize(it,ptr);

    ptr->player=Serializer::deserializeChar(it);
    ptr->ship=Serializer::deserializeChar(it);

    ptr->live=Serializer::deserializeFloat(it);
    ptr->nodamageCountdown=Serializer::deserializeFloat(it);

    FOR(q,WEAPON_SLOTS) {
        delete ptr->weaponSlots[q];
        ptr->weaponSlots[q]=WeaponSlot::_deserialize(it);

    }
    return ptr;
}
// }}}

// {{{ Ship
Ship::Ship(){
    assert(initialized==1,""); 
    radius=defaultRadius;
    mass=defaultMass;
    angularMass=defaultRotationalMass;
    nodamageCountdown=defaultNodamageTime;
    live=defaultLive;
    lastDamageFrom=0;
    lastDamage=0;

    assert(WEAPON_SLOTS==3,"xx");
    weaponSlots[0]=new PlasmaWeaponSlot();
    weaponSlots[1]=new CannonWeaponSlot();
    weaponSlots[2]=new MissileWeaponSlot();
    
}
// }}}

// {{{ timeNotify
void Ship::timeNotify(double timeDelta){
    MassObject::timeNotify(timeDelta);
    nodamageCountdown-=timeDelta;
    if (nodamageCountdown<0) nodamageCountdown=0;

    FOR(q,WEAPON_SLOTS)
        weaponSlots[q]->timeNotify(timeDelta);
}
// }}}

// {{{ destroy
void Ship::destroy(World* world){
  world->addShip(player,ship);
  Explosion* ptr=world->addExplosion(this->position, 
          explosionStrength+lastDamageFrom*lastDamageExplosionFactor);

  ptr->owner=-1;
  ptr->velocity=velocity;

  world->playerStatistics.addDeath(this->player);
  world->playerStatistics.addKill(this->lastDamageFrom);

  FOR(q,WEAPON_SLOTS){
      delete weaponSlots[q];
      weaponSlots[q]=NULL;
  }
}
// }}}

// {{{ doDamage
void Ship::doDamage(double impactDamage, double explosionDamage, int from,World* world){
    if (isImmutable()) return;

    MassObject::doDamage(impactDamage, explosionDamage,from,world);
    world->playerStatistics.addDamage(from,0,impactDamage+explosionDamage,this->player);
    lastDamage+=impactDamage+explosionDamage;
    lastDamageFrom=from;
    live-=impactDamage+explosionDamage;
    if (live<0) {
        alive=0;

        if (from==-1) {
            if (impactDamage==INFINITE_DAMAGE) {
                world->communication.blackHoleKilledShip(player);
            } else if (explosionDamage>0) {
                world->communication.explosionKilledShip(player);
            } else {
                world->communication.impactKilledShip(player);
            }
        } else {
            world->communication.shipKilledShip(from,player);

        }

    }
}
// }}}

// {{{ initialize
void Ship::initialize(){
    defaultLive=Preferences::_getDouble("/ship/defaultLive");
    defaultNodamageTime=Preferences::_getDouble("/ship/defaultNodamageTime");
    defaultMass=Preferences::_getDouble("/ship/defaultMass");
    defaultRadius=Preferences::_getDouble("/ship/defaultRadius");
    defaultRotationalMass=Preferences::_getDouble("/ship/defaultRotationalMass");
    maxSpeed=Preferences::_getDouble("/ship/maxSpeed");
    maxAcceleration=Preferences::_getDouble("/ship/maxAcceleration");
    maxRotationalAcceleration=Preferences::_getDouble("/ship/maxRotationalAcceleration");

    explosionStrength=Preferences::_getDouble("/ship/explosion");
    lastDamageExplosionFactor=Preferences::_getDouble("/ship/lastDamageExplosionFactor");
    autoDestructLastDamage=Preferences::_getDouble("/ship/autoDestructLastDamage");


    PlasmaWeaponSlot::initialize();
    CannonWeaponSlot::initialize();
    MissileWeaponSlot::initialize();
    initialized=1;
}
// }}}

// {{{ autoDestruct
void Ship::autoDestruct(World* world){
    alive=false;
    lastDamage=autoDestructLastDamage;
    world->playerStatistics.addDamage(player,0,live,player);
    lastDamageFrom=owner; // selfkill
}
// }}}

// {{{ doPlayerMove
void Ship::doPlayerMove(World* world){
    Point direction=Point(cos(angle),sin(angle));
    double timeDelta=getTimeDelta();
    doImpact(direction*response.accelerationJets*
            timeDelta*maxAcceleration,
            response.rotationJets*timeDelta*
            maxRotationalAcceleration);

    if (response.fireWeapon>=0 && response.fireWeapon<WEAPON_SLOTS)
        weaponSlots[response.fireWeapon]->shoot(
                this,response.weaponAngle,response.weaponSpeed,
                response.weaponAdditional,world);
    if (response.fireWeapon==-2) autoDestruct(world);
}
// }}}

// {{{ draw
void Ship::draw(){
    ObjectGraphics::drawShip(this);

}
// }}}
