#include "Utils.h"
#include "MyMath.h"
#include <cmath>
#include "Errors.h"
#include "MyStringUtils.h"
#include "Random.h"

#define __ROTATIONAL_CORRECTION 0.97
#define __ACCELERATION_CORRECTION 1.02
#define __ROTATION_TARGET_WEIGHT 0.0
#define __EFFICIENCY 0.3
#define __HALF_JETS_DISTANCE 20

// {{{ shipNavigate
void Utils::shipNavigate(Ship* ship, PlayerResponse* response, Point
        target){
     double a=getShipAccelerationJets(ship,target);
     response->accelerationJets=a;
     int isAccelerating=(a>=0);

     double
         angle=getObjectToTargetAngleWithRespectToVelocity(ship,target,isAccelerating,1);
     response->rotationJets=getShipRotationalJets(ship,angle);
 }
// }}}
 
// {{{ shipNavigate
void Utils::shipNavigate(Ship& ship, PlayerResponse& response,
         Point target){
     return shipNavigate(&ship,&response,target);

 }
// }}}

// {{{ getMassObjectRotationalJets
double Utils::getMassObjectRotationalJets(MassObject* ship, double angle,double acceleration){
  double angleDelta=angle-ship->angle;
  double jets=1;

  angleDelta=normalizeAngle(angleDelta);


  double velocity=ship->angularVelocity*__ROTATIONAL_CORRECTION;

  double stopTime=std::abs(velocity/(acceleration*jets));
  double stopAngle=1/2.0 * acceleration*jets * stopTime* stopTime *
      MyMath::signum(velocity);

  double stop=angleDelta-stopAngle;

  return jets*MyMath::signum(normalizeAngle(stop));
}
// }}}

// {{{ getObjectToTargetAngleWithRespectToVelocity
double Utils::getObjectToTargetAngleWithRespectToVelocity(MassObject*
        object, Point target,int wantToAccelerate, double weight){

   

    int smer=wantToAccelerate==1?1:-1;

    Point d=target-object->position;
    Point v=object->velocity;

    //return d.angle();
    if (v.length()<1e-4) return d.angle();

    double change=(d^v)/d.length()/v.length();

    return d.angle()-change*smer;

}
// }}}

// {{{ getShipRotationalJets
double Utils::getShipRotationalJets(Ship* ship, double angle){
    return
        getMassObjectRotationalJets(ship,angle,Ship::maxRotationalAcceleration);
}
// }}}

// {{{ getShipAccelerationJets
double Utils::getShipAccelerationJets(Ship* ship, Point target){
  double acceleration=Ship::maxAcceleration;
  double jets=1;
  Point direction=target - ship->position;

  Point shipDirection=Point(ship->angle);

  //double
  //    efficiency=(direction*shipDirection)/direction.length()/shipDirection.length();

  double distance=(direction*shipDirection)*__ACCELERATION_CORRECTION;
  if (distance<__HALF_JETS_DISTANCE) jets=0.5;
  if (std::abs(distance)<1) distance+=Random::getDouble(-0.5,0.5);


  //efficiency=1; //sqrt(std::abs(efficiency));

  double velocity=ship->velocity.length()*__ACCELERATION_CORRECTION;
  velocity*=(double) MyMath::signum(shipDirection*ship->velocity);
  

  double stopTime=std::abs(velocity/(acceleration*jets));

  double stopDistance=1/2.0 * acceleration*jets * stopTime* stopTime *
      MyMath::signum(velocity);

  double stop=distance-stopDistance;

  return jets*MyMath::signum(stop);
}
// }}}
 

// {{{ normalizeAngle
double Utils::normalizeAngle(double angle){
    if (std::abs(angle)>10e9) return 0; // upps, something wrong
    int t=(int) (angle/2/M_PI);
    angle-=t*M_PI*2;
    while (angle>M_PI) angle-=2*M_PI;
    while (angle<=-M_PI) angle+=2*M_PI;
    return angle;
}
// }}}

// {{{ distance
double Utils::distance(BasicObject *a, BasicObject *b){
    return (a->position-b->position).length();
}
// }}}

// {{{ getAngle
double Utils::getAngle(Point observer, Point target){
    Point p=target-observer;
    return atan2(p.y,p.x);
}
// }}}

// {{{ getAngle
double Utils::getAngle(Point a,Point b,Point c){
    return normalizeAngle(getAngle(b,c)-getAngle(b,a));
}
// }}}

// {{{ getAngle
double Utils::getAngle(BasicObject *observer, BasicObject* target){
    return getAngle(observer->position,target->position);
}
// }}}

// {{{ getRelativeAngle
double Utils::getRelativeAngle(BasicObject* observer, BasicObject*
        target){
    return normalizeAngle(getAngle(observer,target)-observer->angle);
}
// }}}


// {{{ getPosition
Point Utils::getPosition(Point position, Point velocity, Point
        acceleration, double time){
    Point res=position;
    res+=velocity*time;

    res+=acceleration*time*time*0.5;

    return res;
}
// }}}


// {{{ getPosition
Point Utils::getPosition(MassObject* object, Point acceleration,
        double time){
    return
        getPosition(object->position,object->velocity,acceleration,time);
}
// }}}

// {{{ getPosition
Point  Utils::getPosition(MassObject* object, double time){
    return getPosition(object,Point(),time);
}
// }}}

// {{{ getMinimalDistance
DistanceTime Utils::getMinimalDistance(Point pos1, Point vel1, Point pos2,
        Point vel2){
    DistanceTime result;

    Point p=pos1-pos2;
    Point d=vel1-vel2;


    double t=0;
    if (d.length2()>1e-6) {
        t=-(p*d)/d.length2();
    };

    if (t<0) t=0;

    result.distance=(p+t*d).length();
    result.time=t;
    return result;
}
// }}}

// {{{ getMinimalDistance
DistanceTime Utils::getMinimalDistance(MassObject* object1, MassObject*
        object2){
    return getMinimalDistance(object1->position,object1->velocity,
                              object2->position,object2->velocity);
}
// }}}



double Utils::getTimeToDistance(Point pos1, Point vel1, Point pos2,
        Point vel2, double distance){
    DistanceTime tmp=getMinimalDistance(pos1,vel1,pos2,vel2);
    if (tmp.distance > distance) return -1;
    if (tmp.time==0) return 0;

    Point p=pos1-pos2;
    Point v=vel1-vel2;
    if (p.length()<=distance) return 0;
    double c=p^v;

    double diskriminant=-c*c+MyMath::sqr(distance)*v.length2();
    assert(diskriminant>=0,"whoops - diskriminant is negative!");
    
    double t=-((p*v)+sqrt(diskriminant))/v.length2();
    assert(t>=0,"whoops - why is t smaller than zero?");
    return t;

}

double Utils::getTimeToDistance(MassObject* object1, MassObject*
        object2, double distance){
    return getTimeToDistance(object1->position, object1->velocity,
                            object2->position,
                            object2->velocity,distance);
}
        
