#include "World.h"
#include "Macro.h"
#include "Errors.h"
#include "Assert.h"
#include "Debug.h"
#include "MyStringUtils.h"
#include "Preferences.h"
#include "Wiper.h"
#include "Serializer.h"
#include "BlackHole.h"
#include "CollisionSolver.h"
#include "PlasmaBall.h"
#include "CannonBall.h"
#include "Types.h"
#include "Twirler.h"

#include <algorithm>
#include <stdlib.h>
#include <cmath>

#include "Proboj.h"
#include "Ship.h"


//TODO
// {{{ initializeServer
void World::initializeServer(){
    communication.initializeServer();
    Errors::_addError("waiting for players\n",Errors::NOTICE);
    while (communication.playersConnected()<
            (int)ships.size());
    Errors::_addError("done, starting ...",Errors::NOTICE);


}
// }}}

//TODO
// {{{ prepareDataForPlayer
void World::prepareDataForPlayer(int player,WorldArrays*
        arrays){
    arrays->clearAll();

    arrays->actualTime=time;
    FOREACH(it,ships) {
        if ((*it)->player==player) {
            arrays->myShips.push_back(*(*it));
        } else 
            if ((*it)->visible)  {
            Ship tmp=*(*it);
            tmp.clearForOthers();
            arrays->otherShips.push_back(tmp);
        }
    }


    FOREACH(it,heavyObjects){
        BlackHole* ptr=dynamic_cast<BlackHole*> (*it);
        assert(ptr!=NULL,"oops, dynamic cast failed");
        arrays->blackHoles.push_back(*ptr);
    }


    FOREACH(it, lightObjects){
        int type=(*it)->getType();
        if (!(*it)->visible) continue;

        switch(type){
            case Types::asteroid:{
                Asteroid* ptr1=dynamic_cast<Asteroid*> (*it);
                assert(ptr1!=NULL,"oops, dynamic_cast failed");
                arrays->asteroids.push_back(*ptr1);
                break;
	    }

            case Types::plasmaBall:{
                PlasmaBall* ptr2=dynamic_cast<PlasmaBall*> (*it);
                assert(ptr2!=NULL,"oops, dynamic_cast failed");
                arrays->plasmaBalls.push_back(*ptr2);
                break;
       	   }

            case Types::cannonBall:{
                CannonBall* ptr3=dynamic_cast<CannonBall*> (*it);
                assert(ptr3!=NULL,"oops, dynamic_cast failed");
                arrays->cannonBalls.push_back(*ptr3);
                break;
	    }
	    
	    case Types::missile:{
                Missile* ptr3=dynamic_cast<Missile*> (*it);
                assert(ptr3!=NULL,"oops, dynamic_cast failed");
                arrays->missile.push_back(*ptr3);
                break;
	    }


            default:
                    assert(false,"whoops, unknown light object");
        }


    }

    FOREACH(it, masslessObjects){
        int type=(*it)->getType();
        assert((*it)->visible==true,"affector not visible?");

        switch(type){
            case Types::explosion:{
                Explosion* ptr1=dynamic_cast<Explosion*> (*it);
                assert(ptr1!=NULL,"oops, dynamic_cast failed");
                arrays->explosions.push_back(*ptr1);
                break;
	    }
            case Types::nebula:{
                Nebula* ptr2=dynamic_cast<Nebula*> (*it);
                assert(ptr2!=NULL,"oops, dynamic_cast failed");
                arrays->nebulas.push_back(*ptr2);
                break;
           }

            case Types::twirler:{
                Twirler* ptr3=dynamic_cast<Twirler*> (*it);
                assert(ptr3!=NULL,"oops, dynamic_cast failed");
                arrays->twirlers.push_back(*ptr3);
                break;
	    }

            default:
                    assert(false,"whoops, unknown affector");
        }


    }




}

// }}}

// {{{ addLightObject
void World::addLightObject(MassObject* ptr){
    lightObjects.push_back(ptr);
}
// }}}

// {{{ addShip OK
Ship* World::addShip(int player, int ship){
  Ship* ptr=new Ship();
  ptr->position=collider.getSpawnPoint();
  ptr->player=player;
  ptr->owner=player;
  ptr->ship=ship;
  ships.push_back(ptr);
  return ptr;
}
// }}}

// {{{ addAffector
void World::addAffector(Affector* ptr){
    masslessObjects.push_back(ptr);
}
// }}}

// {{{ addExplosion OK
Explosion* World::addExplosion(Point position,double damagePerSecond){
    Explosion* ptr=new Explosion();
    ptr->position=position;
    ptr->damagePerSecond=damagePerSecond;
    masslessObjects.push_back(ptr);
    communication.explode();
    return ptr;
}
// }}}


// {{{  loadBlackHoles OK
void World::loadBlackHoles(Preferences& mapData){
  int blackHoles=mapData.getInt("/blackHole/count");
  FOR(q,blackHoles){
      std::string prefix="/blackHole/" +
          MyStringUtils::intToString(q);

      Point position=Point::getPointFromString(
              mapData.getString(prefix+"/position"));
      Point velocity=Point::getPointFromString(
              mapData.getString(prefix+"/velocity"));
      double mass=mapData.getDouble(prefix+"/mass");
      double radius=mapData.getDouble(prefix+"/radius");

      BlackHole* object=new BlackHole();
      object->mass=mass;
      object->position=position;
      object->velocity=velocity;
      object->radius=radius;
      object->angle=0;
      object->angularMass=mass*radius*radius/2;
      heavyObjects.push_back(object);
  }
}
// }}}

// {{{ loadTwirlers
void World::loadTwirlers(Preferences& mapData){
    int twirlers=mapData.getInt("/twirler/count");
    FOR(q, twirlers){
        std::string prefix="/twirler/" +
            MyStringUtils::intToString(q);

        Point position=Point::getPointFromString(
                mapData.getString(prefix+"/position"));
        double radius = mapData.getDouble(prefix+"/radius");
        double maxSpin= mapData.getDouble(prefix+"/maxSpin");
        double spinup = mapData.getDouble(prefix+"/spinup");

        Twirler* object=new Twirler();
        object->position=position;
        object->maxSpin=maxSpin;
        object->radius=radius;
        object->spinPerSecond=spinup;
        object->clockwise=0;
        addAffector(object);

    }
}
// }}}

// {{{ loadNebulas
void World::loadNebulas(Preferences& mapData){
    int nebulas=mapData.getInt("/nebula/count");
    FOR(q,nebulas){
        std::string prefix="/nebula/" +
            MyStringUtils::intToString(q);

        Point position=Point::getPointFromString(
                mapData.getString(prefix+"/position"));
        double radius = mapData.getDouble(prefix+"/radius");

        Nebula* ptr=new Nebula();
        ptr->position=position;
        ptr->radius=radius;
        addAffector(ptr);
    }
}
// }}}


// {{{ loadAffectors
void World::loadAffectors(Preferences& mapData){
    loadTwirlers(mapData);
    loadNebulas(mapData);
}
// }}}

// {{{ loadShips OK
void World::loadShips(Preferences& mapData){
    shipCount.clear();

 FOR(q,MAX_PLAYERS) {
     int shipss = mapData.getInt("/ship/player" +
             MyStringUtils::intToString(q) + "/count");
     shipCount.push_back(shipss);
     FOR(w,shipss) {
         addShip(q,w);
    }
 }
}
// }}}

// {{{ draw OK
void World::draw(){
   
    FOREACH(it,heavyObjects)
         (*it)->draw();
    FOREACH(it,lightObjects)
         (*it)->draw();
    FOREACH(it,ships)
         (*it)->draw();
    FOREACH(it,masslessObjects)
         (*it)->draw();
}
// }}}

// {{{ getObjectCount OK
int World::getObjectCount(){
    return heavyObjects.size()+
           lightObjects.size()+
           ships.size()+
           masslessObjects.size();
}
// }}}

// TODO
// {{{ initialize
void World::initialize(const std::string& mapFile){
    DEBUG("initialize: "+mapFile);
  clearAllObjects();
  Ship::initialize();
  Explosion::initialize();
  PlasmaBall::initialize();
  CannonBall::initialize();
  Missile::initialize();
  Asteroid::initialize();

  Preferences mapData;
  mapData.loadFromFile(mapFile);

  mapName=mapData.getString("/map/name");
  width=mapData.getInt("/world/width");
  height=mapData.getInt("/world/height");

  collider.initialize(this);
  CollisionSolver::initialize();
  edgeCollider.initialize(width,height);
  gravity.initialize();
  time=0;
  roundTime=mapData.getDouble("/world/roundTime");


  maxFrameTime=Preferences::_getDouble("/world/maxFrameTime");

  loadBlackHoles(mapData);
  loadShips(mapData);
  asteroidSpawningSubsystem.init(mapData);
  loadAffectors(mapData);

  communication.sendNewGame(&communicationGraph,shipCount);
  playerStatistics.initialize(MAX_PLAYERS);
}
// }}}

// {{{ getRemainingTime() OK
double World::getRemainingTime(){
    return roundTime-time;
}
// }}}

// {{{ World OK
World::World():frame(0),maxFrameTime(1){
    Serializer::checkTypes();
}
// }}}

// {{{ ~World OK
World::~World(){
 clearAllObjects();
}
// }}}

// {{{ serialize OK
void World::serialize(std::vector<unsigned char>& destination){
    Serializer::serializeInt(destination,frame);
    Serializer::serializeFloat(destination,time);
    Serializer::serializeInt(destination,width);
    Serializer::serializeInt(destination,height);
    Serializer::serializeObjects(destination,heavyObjects);
    Serializer::serializeObjects(destination,lightObjects);
    Serializer::serializeObjects(destination,ships);
    Serializer::serializeObjects(destination,masslessObjects);
}
// }}}

// {{{ deserialize OK
void World::deserialize(std::vector<unsigned char>::iterator& it){
 clearAllObjects();
 frame=Serializer::deserializeInt(it);
 time=Serializer::deserializeFloat(it);
 width=Serializer::deserializeInt(it);
 height=Serializer::deserializeInt(it);
 Serializer::deserializeObjects(it,heavyObjects);
 Serializer::deserializeObjects(it,lightObjects);
 Serializer::deserializeObjects(it,ships);
 Serializer::deserializeObjects(it,masslessObjects);
}
// }}}

// {{{ computePlayerResponses OK
void World::computePlayerResponses(double timeDelta){
    (void) timeDelta;
    FOREACH(it,ships)
        (*it)->doPlayerMove(this);
}
// }}}

// {{{ computeNextFrame OK
void World::computeNextFrame(double timeDelta){
//    DEBUG("next frame");
    time+=timeDelta;

    if (timeDelta>maxFrameTime) printf("FRAME SKIP - TIME %lf\n",timeDelta);
    while (timeDelta > 0){
        double td;
        if (timeDelta > maxFrameTime) {
            td = maxFrameTime;
            Errors::_addError("World: big time skip, performance "
                    "problems?",Errors::WARNING);
        } else { 
            td = timeDelta;
        } 

//        DEBUG("xx");
        nextFrame(td);
        timeDelta -= td;
    }
    ++frame;
}
// }}}

// {{{ nextFrame OK
void World::nextFrame(const double timeDelta){
    while (asteroidSpawningSubsystem.getSpawnCount()>0)
        lightObjects.push_back(
                asteroidSpawningSubsystem.spawn(
                    collider.getSpawnPoint()));

    clean();
    FOREACH(it,ships) (*it)->timeNotify(timeDelta);
    computePlayerResponses(timeDelta);
    timeNotify(timeDelta);
    computeGravitationalAcceleration();
    computeCollisions();
    computeEdgeCollisions();
    flip();
    wipe();
}
// }}}

// {{{ clean OK
void World::clean(){
    FOREACH(it, heavyObjects) (*it)->clean();
    FOREACH(it, lightObjects) (*it)->clean();
    FOREACH(it, ships)        (*it)->clean(); 
    FOREACH(it, masslessObjects) (*it)->clean(); 
}
// }}}

// {{{ clearAllObjects OK
void World::clearAllObjects(){
    FOREACH(it,heavyObjects) delete (*it);
    heavyObjects.clear();
    FOREACH(it,lightObjects) delete (*it);
    lightObjects.clear();
    FOREACH(it,masslessObjects) delete (*it);
    masslessObjects.clear();
    FOREACH(it,ships) delete (*it);
    ships.clear();
}
// }}}

// {{{ timeNotify OK
void World::timeNotify(double timeDelta){
    asteroidSpawningSubsystem.updateTime(timeDelta);
    FOREACH(it, heavyObjects)    (*it)->timeNotify(timeDelta);
    FOREACH(it, lightObjects)    (*it)->timeNotify(timeDelta);
    FOREACH(it, ships)           (*it)->timeNotify(timeDelta);
    FOREACH(it, masslessObjects) (*it)->timeNotify(timeDelta);
}
// }}}

// {{{ flip OK
void World::flip(){
    FOREACH(it, heavyObjects) (*it)->flip();
    FOREACH(it, lightObjects) (*it)->flip();
    FOREACH(it, ships)        (*it)->flip();
}
// }}}

// {{{ computeEdgeCollisions OK
void World::computeEdgeCollisions(){
    edgeCollider.collideMassObjects(heavyObjects.begin(),
            heavyObjects.end());
    edgeCollider.edgeKillMassObjects(lightObjects.begin(),
            lightObjects.end());
    edgeCollider.collideMassObjects(ships.begin(),ships.end());
}
// }}}

// {{{ computeCollisions OK
void World::computeCollisions(){
    collider.clear();
    collider.insertAndCollideMassObjects(heavyObjects.begin(),heavyObjects.end());
    collider.insertAndCollideMassObjects(lightObjects.begin(),lightObjects.end());
    collider.insertAndCollideMassObjects(ships.begin(),ships.end());

    collider.collideMasslessObjects(masslessObjects.begin(),masslessObjects.end());
}
// }}}

// {{{ computeGravitationalAcceleration OK
void World::computeGravitationalAcceleration(){
    gravity.clear();
    gravity.setGravitySources(heavyObjects.begin(),
            heavyObjects.end());

    gravity.applyGravityTo(heavyObjects.begin(), heavyObjects.end());
    gravity.applyGravityTo(lightObjects.begin(), lightObjects.end());
    gravity.applyGravityTo(ships.begin(),        ships.end());
}
// }}}

// {{{ wipe OK
void World::wipe(){
    Wiper::wipeMassObjects(heavyObjects,this);
    Wiper::wipeMassObjects(lightObjects,this);
    Wiper::wipeMassObjects(ships,this);
    Wiper::wipeMasslessObjects(masslessObjects,this);
}
// }}}

// {{{ receiveDataFromPlayers OK
void World::receiveDataFromPlayers(){
 FOREACH(it,ships){
     PlayerResponse* response;
     response=communication.receivePlayerResponse((*it)->player,(*it)->ship);
     if (response!=NULL) {
         (*it)->response=(*response);
     }
     delete response;
 }
}
// }}}

// TODO
// {{{ sendDataToViewer
void World::sendDataToViewer(){
    communication.sendDataToViewer(this);	
}
// }}}

// {{{ sendDataToStats
void World::sendDataToStats(){
    communication.sendStatisticsData(&playerStatistics);
}
// }}}

// {{{ sendDataToPlayers OK
void World::sendDataToPlayers(){
   communication.sendPlayerData(this);
}
// }}}  

